from .utilites import print_color
from functools import wraps


class PyMikroTikExceptions(Exception):
    pass


class ConnectError(PyMikroTikExceptions):
    pass


class IpAddressFormatError(PyMikroTikExceptions):
    pass


class RouterError(PyMikroTikExceptions):
    pass


class InvalidSearchAttribute(PyMikroTikExceptions):
    pass


def exception_control(method):
    @wraps(method)
    def wrapper(*args, **kwargs):
        if args[0].connection._ignore_errors is False:
            return method(*args, **kwargs)
        try:
            return method(*args, **kwargs)
        except PyMikroTikExceptions as err:
            print_color(str(err), 'red')
            return err
    return wrapper



import paramiko
from .ip import Ip
from .exceptions import ConnectError
from .utilites import print_color


class MikroTikConnect:
    """Main router management class"""
    def __init__(self, host: str, login: str, password: str, port: int = 22,
                 auto_connection: bool = False,
                 ignore_errors: bool = False):
        self.host = host
        self.port = port
        self.__password = password
        self.__login = login
        self._auto_connection = auto_connection
        self._ignore_errors = ignore_errors
        self.__client = paramiko.SSHClient()
        self.__client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    def connect(self):
        """Connect to the router"""
        try:
            self.__client.connect(hostname=self.host, port=self.port, username=self.__login, password=self.__password)
            print_color(f'Successful connection to {self.host}', 'cyan')
        except paramiko.ssh_exception.AuthenticationException:
            raise ConnectError('Ошибка аутентификации. Не верный логин или пароль')
        except paramiko.ssh_exception.NoValidConnectionsError:
            raise ConnectError(f'Не возможно подключится к хосту {self.host} по порту {self.port}')

    def disconnect(self):
        self.__client.close()

    @property
    def ip(self):
        """
        ip
        :return: class Ip
        """
        return Ip(connection=self)

    @staticmethod
    def __check_connection(method):
        def wrapper(*args, **kwargs):
            if args[0]._auto_connection is False:
                return method(*args, **kwargs)
            try:
                funk = method(*args, **kwargs)
                return funk
            except AttributeError:
                args[0].connect()
                return method(*args, **kwargs)
        return wrapper

    @__check_connection
    def send_command(self, command) -> str:
        """Accepts the command. Returns the response"""
        _, stdout, __ = self.__client.exec_command(command)
        output = stdout.read().decode('utf-8')
        return output




from .exceptions import IpAddressFormatError, InvalidSearchAttribute
from abc import ABC, abstractmethod


class MTObject(ABC):
    @abstractmethod
    def get_one(self, find_by: dict):
        pass

    @abstractmethod
    def get_all(self) -> list:
        pass

    def _checking_find_by_field(self, find_by: dict) -> bool:
        """Возращает True если все поля поиска существуют для данного объекта.
        Иначе генерируется исключение InvalidSearchAttribute"""
        if not isinstance(find_by, dict):
            raise ValueError("'find_by' must by 'dict'")
        attributes_ls = list(vars(self).keys())
        if self.__class__.__name__ == 'Address':
            attributes_ls.remove('_numbers')
        attributes_ls.remove('connection')
        attributes_ls.remove('flags')
        for field in find_by.keys():
            if field not in attributes_ls:
                raise InvalidSearchAttribute(f'The "{field}" attribute does not exist for the object. \n'
                                             f'Use only {attributes_ls}')
        return True


CIDR = {
    '255.255.255.255': 32, '255.255.255.254': 31, '255.255.255.252': 30, '255.255.255.248': 29,
    '255.255.255.240': 28, '255.255.255.224': 27, '255.255.255.192': 26, '255.255.255.128': 25,
    '255.255.255.0': 24, '255.255.254.0': 23, '255.255.252.0': 22, '255.255.248.0': 21,
    '255.255.240.0': 20, '255.255.224.0': 19, '255.255.192.0': 18, '255.255.128.0': 17,
    '255.255.0.0': 16, '255.254.0.0': 15, '255.252.0.0': 14, '255.248.0.0': 13,
    '255.240.0.0': 12, '255.224.0.0': 11, '255.192.0.0': 10, '255.128.0.0': 9,
    '255.0.0.0': 8, '254.0.0.0': 7, '252.0.0.0': 6, '248.0.0.0': 5,
    '240.0.0.0': 4, '224.0.0.0': 3, '192.0.0.0': 2, '128.0.0.0': 1, '0.0.0.0': 0
}


class IPAddress:
    def __init__(self, value: str):
        if not isinstance(value, str):
            raise ValueError('Object must be a string')
        self._value = value.strip()
        self.address = None
        self.mask = None
        self.__checking_value()

    def __checking_value(self):
        address_mask = self._value.split('/')
        try:
            if len(address_mask) > 2:
                raise ValueError
            address = address_mask[0].split('.')
            if len(address) != 4:
                raise ValueError
            for a in address:
                _ip = int(a)
                if _ip < 0 or _ip > 255:
                    raise ValueError
            self.address = address_mask[0]
            if len(address_mask) == 2:
                if address_mask[1] in CIDR.keys():
                    self.mask = CIDR[address_mask[1]]
                else:
                    _mask = int(address_mask[1])
                    if _mask in CIDR.values():
                        self.mask = str(address_mask[1])
                    else:
                        raise ValueError
                if self.mask == '0' and self.address != '0.0.0.0':
                    raise IpAddressFormatError('In this case, the subnet mask cannot be 0.')
        except ValueError:
            raise IpAddressFormatError('Invalid IP address format.\n'
                                       'Example: 192.168.32.15/24 or 192.168.32.15/255.255.255.0')

    def __str__(self):
        address = self.address
        if self.mask:
            address += f'/{self.mask}'
        return address





from colorama import Fore, init
init(autoreset=True)


def print_color(text: str, color: str | None = None):
    COLORS = {
        'cyan': Fore.CYAN,
        'red': Fore.RED,
        'magenta': Fore.MAGENTA,
        'yellow': Fore.YELLOW,
        'blue': Fore.BLUE
    }

    if not color:
        color = 'not_color'

    print(COLORS.get(color, Fore.RESET) + text)




from .router_objects import MTObject, IPAddress
from .exceptions import exception_control, RouterError, InvalidSearchAttribute
from .utilites import print_color
import re


class Address(MTObject):
    """
    Address managemen
    /ip address (object)
    """
    def __init__(self, connection):
        self.connection = connection
        self.address: IPAddress | None = None
        self.network: IPAddress | None = None
        self.interface: str | None = None
        self.comment: str | None = None
        self.numbers: str | None = None

        self.disabled: bool | None = None
        self.flags: str | None = None

    @exception_control
    def add(self, address: str, interface: str, network: str = None,
            comment: str = None, disabled: bool = False) -> 'Address':
        """
        Create a new item

        :param address: Local IP address
        :param interface: Interface name
        :param network: Network
        :param comment: Short description of the item
        :param disabled: Disable items
        :return: 'Address'
        """
        new_address = Address(connection=self.connection)
        new_address.address = IPAddress(address)
        if network:
            new_address.network = IPAddress(network)
        new_address.interface = interface
        new_address.comment = comment
        new_address.disabled = disabled
        command = f'/ip address add address="{new_address.address}" interface="{interface}"'
        command += f' network="{new_address.network}"' if network else ''
        command += f' comment="{new_address.comment}"' if comment else ''
        command += f' disabled="yes"' if disabled is True else ''

        result = self.connection.send_command(command)
        if result:
            raise RouterError(result)
        print_color(f"IP address '{new_address.address}' added to interface '{new_address.interface}'", 'cyan')
        return new_address

    @exception_control
    def set(self, find_by: dict or int, address: str = None, interface: str = None,
            network: str = None, disabled: bool = None, comment: str = None) -> 'Address':
        """
        Change item properties

        :param find_by: Either a dictionary or an integer indicating how to look up the IP address.
                        If it is a dictionary, it must contain key-value pairs for identifying the IP address.
                        If it is an integer, it represents a number from a list of IP addresses.
        :param address: Local IP address
        :param interface: Interface name
        :param network: Network
        :param disabled: Disable items
        :param comment: Short description of the item
        :return: 'Address'
        """
        edit_address = Address(connection=self.connection)
        if isinstance(find_by, int):
            command = f'/ip address set {find_by}'
        elif find_by.get('numbers', None) is not None:
            command = f'/ip address set {int(find_by["numbers"])}'
        else:
            self._checking_find_by_field(find_by)
            find = '[find'
            for k, v in find_by.items():
                find += f' {k}="{v}"'
            find += ']'
            command = f'/ip address set {find}'
        if address:
            edit_address.address = IPAddress(address)
            command += f' address="{edit_address.address}"'
        if interface:
            edit_address.interface = interface
            command += f' interface="{edit_address.interface}"'
        if network:
            edit_address.network = IPAddress(network)
            command += f' network="{edit_address.network}"'
        if disabled:
            edit_address.disabled = disabled
            command += f' disabled="yes"'
        if comment:
            edit_address.comment = comment
            command += f' comment="{edit_address.comment}"'
        result = self.connection.send_command(command)
        if result:
            raise RouterError(result)
        print_color('IP address parameters changed', 'cyan')
        return edit_address

    def __functions(self, func: str, numbers: int) -> 'Address':
        addresses = self.get_all()
        addresses_numbers = [address.numbers for address in addresses]
        if numbers not in addresses_numbers:
            raise InvalidSearchAttribute(f'There is no number {numbers} in the address list.\n'
                                         f'Here is a list of available numbers\n'
                                         f'{addresses_numbers}')
        command = f'/ip address {func} {numbers}'
        result = self.connection.send_command(command)
        if result:
            raise RouterError(result)
        return self.get_all()[numbers]

    def disable(self, numbers: int) -> 'Address':
        """
        Disable items

        :param numbers: List of item numbers
        :return: 'Address'
        """
        address = self.__functions('disable', numbers)
        print_color(f'IP address {str(address)} disabled', 'cyan')
        return address

    def enable(self, numbers: int):
        """
        Enable items

        :param numbers: List of item numbers
        :return: 'Address'
        """
        address = self.__functions('enable', numbers)
        print_color(f'IP address {str(address)} enabled', 'cyan')
        return address

    def remove(self, numbers: int):
        """
        Remove item

        :param numbers: List of item numbers
        :return: 'Address'
        """
        address = self.__functions('remove', numbers)
        print_color(f'IP address {str(address)} deleted', 'cyan')
        return address

    def get_all(self) -> list['Address']:
        """
        Get all items

        :return: list['Address']
        """
        command = '/ip address print detail'
        result = self.connection.send_command(command)
        pattern = re.compile(r'\s*(\d+)\s+([DIX]?)\s*(?:\s*;;; (.*?))?\s*address=([0-9./]+)\s+network=([0-9.]+)\s+interface=([^\s]+)')
        addresses_string = pattern.findall(result)

        addresses: list[Address] = []
        for _address in addresses_string:
            _index, flag, comment, address, network, interface = _address
            addresses.append(self.__create_an_object(_index, flag, comment, address, network, interface))
        return addresses

    def get_one(self, find_by: dict):
        """
        Get one items

        :param find_by: Either a dictionary or an integer indicating how to look up the IP address.
                        If it is a dictionary, it must contain key-value pairs for identifying the IP address.
                        If it is an integer, it represents a number from a list of IP addresses.
        """
        if isinstance(find_by, int):
            try:
                item = self.get_all()[0]
                return item
            except IndexError:
                raise InvalidSearchAttribute('Not find')
        else:
            self._checking_find_by_field(find_by)
            command = '/ip address print detail where '
            for k, v in find_by.items():
                if v:
                    command += f'{k}="{v}" '
            result = self.connection.send_command(command)
            pattern = re.compile(
                r'\s*(\d+)\s+([DIX]?)\s*(?:\s*;;; (.*?))?\s*address=([0-9./]+)\s+network=([0-9.]+)\s+interface=([^\s]+)')
            try:
                address_ = pattern.findall(result)[0]
            except IndexError:
                raise InvalidSearchAttribute('Not find')
            address = self.__create_an_object(*address_)
            return address

    def print_all(self) -> str:
        """
        Print out all items

        :return: String listing all items
        """
        addresses = self.get_all()
        text = ''
        for address in addresses:
            text += f'\n{str(address)}'
        print_color(text, 'cyan')
        return text

    def __create_an_object(self, _index, flag, comment, address, network, interface):
        ip_address = Address(connection=self.connection)
        ip_address.address = IPAddress(address)
        ip_address.network = IPAddress(network)
        ip_address.interface = interface
        ip_address.numbers = int(_index)
        ip_address.flags = flag
        if flag == 'X':
            ip_address.disabled = True
        if comment:
            ip_address.comment = comment
        return ip_address

    def __str__(self):
        st = f'{self.numbers} ' if self.numbers is not None else ''
        st += f'{self.flags} address={self.address} network={self.network} interface={self.interface}'
        if self.comment:
            st += f' comment={self.comment}'
        return st


class Ip:
    def __init__(self, connection):
        self.connection = connection

    @property
    def address(self):
        address = Address(connection=self.connection)
        return address






from setuptools import setup, find_packages


"""
:authors: Fialkofskiy
:copyright: (c) 2024 Fialkofskiy
"""

version = '0.1.3'
long_description = '''pyMikroTik is a powerful tool for interacting with MikroTik routers, 
providing convenient and flexible device management through SSH connections. 
Developed in Python, the library offers a straightforward interface for performing various operations, 
including parameter configuration, monitoring, and task automation.'''
description = 'pyMikroTik is a powerful tool for interacting with MikroTik routers'


setup(
    name='pyMikroTik',
    version=version,

    author='fialkofskiy',
    author_email='fialkofskiy@gmail.com',

    long_description=long_description,
    description=description,

    packages=find_packages(),
    install_requires=[
        'bcrypt==4.1.2',
        'cffi==1.16.0',
        'cryptography==42.0.1',
        'paramiko==3.4.0',
        'pycparser==2.21',
        'PyNaCl==1.5.0',
        'colorama==0.4.6'
    ]
)


